#!/bin/bash
# Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
# SPDX-License-Identifier: GPL-3.0-only
# Authors: Sergei Zharko [committer], Pierre-Alain Loizeau
#
# @file   reco_mcbm.sh
# @brief  Script to run the TSA file reconstruction in mCBM
# @since  21.05.2024
# @author Sergei Zharko <s.zharko@gsi.de>
#
# **********************************************************************************************************************
# ***                                      User Manual (v. 0.0.2, 05.07.2024)                                        ***
# **********************************************************************************************************************
#
# 1. Definitions
#   <setup>:
#     An official setup tag, e.g. "mcbm_beam_2024_05_08_nickel".
#   <label>:
#     A common part of name bases of the data-files, e.g. "3105_node8_05_0002"
#     NOTE: <label> is strictly determined at least for a particular setup.
#   <top_dir>:
#     A top "data" directory.
#
# 2. Introduction
#
#   This script aims to provide a generic data reconstruction scenario in mCBM, which includes:
#     - setup files generation:  [--setup]
#     - TSA file unpacking       [--unpack]
#     - reconstruction           [--reco]
#     - main QA                  [--qa]
#     - reconstruciton QA (Alex) [--qa-module]
#
#   Required options:
#     --tsa <path>           Path to the input TSA file
#                             NOTE: must contain a run index as a first integer in its base name.
#   Auxiliary options:
#     -n [--nts] <N_TS>  Number of timeslices to procede
#     --setup, --unpack, --reco, --qa, --qa-module
#     --param-online <path>  Path to the online parameters
#                             NOTE: should be ${VMCWORKDIR}/parameters/online, but local user may have copy for testing
#
# 3. File names involved:
#   <geo>:       Input geometry file:  <top_dir>/<setup>.geo.root
#   <par>:       Input parameter file:  <top_dir>/<setup>.par.root --copy--> <top_dir>/<label>.par.root
#   <tra>:       Dummy transport file: <top_dir>/<setup>.tra.root
#   <digi>:      Digitization output: <top_dir>/<label>.digi.root --ln--> <top_dir>/<label>.raw.root
#   <reco>:      Reconstruction output: <top_dir>/<label>.reco.root --ln--> <top_dir>/<label>.rec.root
#   <qa>:        Main QA output: <top_dir>/<label>.qa.root
#   <qa-module>: Reconstruction module QA: <top_dir>/<label>.rqa.root
#
# 4. Knwon missing features
#
#   - Script and its sub-process is not killed if CTEST is stopped with CTRL+C
#   - Error in reco of run 3105:
# CbmAlgoBuildRawEvents::filterBmon: Bmon digi with wrong address [GetSmType() != 5]. Skip.
#   - LOG to change from INFO to DEBUG in RECO CA step:
#[2024-07-08 10:11:10.301054] [0x00007f96766b4a40] [info]    ca: Hit out of range: iHit = 35, time = 126.389 ms, window = 1
#
#   - LOG to change from INFO to DEBUG in RECO CA step:
# =======  Ca Track finder: process event 17297 ...
# [2024-07-08 10:11:10.375592] [0x00007f96766b4a40] [info]    CA tracker process time slice 127.431 -- 127.431 [ms] with 47 hits
# [2024-07-08 10:11:10.375597] [0x00007f96766b4a40] [info]    Fraction of hits from monster events: 0
# [2024-07-08 10:11:10.375603] [0x00007f96766b4a40] [info]    Thread: 0 from 127.431 ms  to 127.431 ms (delta = 0.00012 ms)
# [2024-07-08 10:11:10.376800] [0x00007f96766b4a40] [info]    CA tracker: time slice finished. Reconstructed 4 tracks with 16 hits. Processed 47 hits in 1 time windows. Reco time 1.21725e-12 s
# Ca Track Finder finished, found 4 tracks
# CA Track Finder: 0.00121725 s/sub-ts
# Tracking performance... done
# End of CA
# CA: N hits used/tot = 46/49
#
#   - LOG to change from INFO to DEBUG in QA step: [root.exe][10:11:24][INFO] CA: N hits used/tot = 17/18
#
# *********************
# ** USER PARAMETERS **
# *********************

# Data directory
DATA_TOP_DIR='./data'

# Subsystem flags
RECO_MVD=0
RECO_STS=1
RECO_TRD=1
RECO_TRD2d=1
RECO_RICH=1
RECO_MUCH=0
RECO_TOF=1
RECO_FSD=0
RECO_PSD=0

# Algorithm flags
RECO_TOFtr=0
RECO_ALI=1
RECO_EvB=1
RECO_CA=1
RECO_PV=1
RECO_QA=0


# ----------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------


# *************************
# ** Parameter selection **
# *************************

# ----- Execution steps
DO_CREATE_SETUP=0
DO_UNPACK=0
DO_RECO=0
DO_QA=0         # Main QA Macro
DO_QA_MODULE=0  # QA Macro by Alexandru
ONLINE_PAR=${VMCWORKDIR}/parameters/online

# ----- Run information
RUN=-1
TSA=
N_TS=-1
while [[ $# -gt 0 ]]; do
  case ${1} in
    --setup )
      DO_CREATE_SETUP=1
      ;;
    --unpack )
      DO_UNPACK=1
      ;;
    --reco )
      DO_RECO=1
      ;;
    --qa )
      DO_QA=1
      ;;
    --qa-module )
      DO_QA_MODULE=1
      ;;
    --tsa )
      TSA=${2}
      ;;
    --param-online )
      ONLINE_PAR=${2}
      ;;
    -n | --nts )
      N_TS=${2}
      ;;
  esac
  shift
done

# ----- Check the environment
#
if [[ -z "${VMCWORKDIR}" ]]; then
  printf "E- CBM environment is not defined (VMCWORKDIR is not found). Please, configure your CbmRoot\n"
  exit 1
fi

# ----- Select the reconstruction binary
ONLINE_BINARY="${VMCWORKDIR}/../../bin/cbmreco"
if [[ ! -x "${ONLINE_BINARY}" ]]; then
  ONLINE_BINARY=$(which cbmreco)
  if [[ ! -x "${ONLINE_BINARY}" ]]; then
    # Test for execution within the build folder by CTEST (CI/CMAKE/CDASH)
    ONLINE_BINARY="${PWD}/../../../bin/cbmreco"
    if [[ ! -x "${ONLINE_BINARY}" ]]; then
      printf "E- Online binary was not found. Please:\n"
      printf "   - install and configure your CbmRoot\n"
      printf "   - or provide the bin directory in the PATH environmental variable\n"
      printf "   - or run within the CI/CTEST environment\n"
      exit 2
    fi
  fi
fi

printf "I- Online binary path: %s, with parameters: %s\n" ${ONLINE_BINARY} ${ONLINE_PAR}

# ----- Check the TSA input and retrieve the run index
if [[ -z ${TSA} ]]; then
  printf "E- TSA input file is not defined\n"
  exit 3
fi
TSA_INFO=($(basename $(echo ${TSA}) | grep -oE '[0-9]*'))

RUN=${TSA_INFO[0]}  # Implying, that TSA basename contains run index as the first integer
if [[ ${RUN} -lt 0 ]]; then
  printf "E- Run number is undefined, please try again with option -r <RUN> or --run <RUN>\n"
fi

mkdir -p ${DATA_TOP_DIR}

# ----- Filenames definition
TSA_INP=${TSA}
MACRO_SETUP="${VMCWORKDIR}/macro/run/create_mcbm_geo_setup.C"
MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2022/mcbm_event_reco_L1.C"
MACRO_QA="${VMCWORKDIR}/macro/mcbm/mcbm_qa.C"
MACRO_QA_MODULE="${VMCWORKDIR}/macro/qa/run_recoQa.C"
SETUP_NAME=

# ----- !!! Setting selections vs.

if [[ ${RUN} -ge 2350 && ${RUN} -le 2397 ]]; then
  SETUP_NAME="mcbm_beam_2022_05_23_nickel"
  MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2022/mcbm_event_reco_L1.C"
elif [[ ${RUN} -ge 2454 && ${RUN} -le 2497 ]]; then
  SETUP_NAME="mcbm_beam_2022_06_16_gold"
  MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2022/mcbm_event_reco_L1.C"
elif [[ ${RUN} -ge 2498 && ${RUN} -le 2610 ]]; then
  SETUP_NAME="mcbm_beam_2022_06_18_gold"
  MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2022/mcbm_event_reco_L1.C"
elif [[ ${RUN} -ge 2724 && ${RUN} -le 2917 ]]; then
  SETUP_NAME="mcbm_beam_2024_03_22_gold"
  MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2024/mcbm_event_reco_L1.C"
elif [[ ${RUN} -ge 2918 ]]; then
  SETUP_NAME="mcbm_beam_2024_05_08_nickel"
  MACRO_RECO="${VMCWORKDIR}/macro/beamtime/mcbm2024/mcbm_event_reco_L1.C"
else
  printf "E- Run %5d is undefined. Exiting" "${RUN}"
  exit 4
fi

# -----
# File
FILE_LABEL=$(printf $(basename ${TSA_INP}) | cut -f 1 -d '.')
OUT_DIGI="${DATA_TOP_DIR}/${FILE_LABEL}.digi.out"
INP_FILE="${DATA_TOP_DIR}/${FILE_LABEL}.digi.root"

# -----
# Log Files
SETUP_LOG="run_mcbm_setup.log"
UNPACK_LOG="run_mcbm_unpack.log"
RECO_LOG="run_mcbm_reco.log"
RECO_QA_LOG="run_mcbm_reco_qa.log"
QA_LOG="run_mcbm_qa.log"

#  *********************
#  ** Steps execution **
#  *********************

# ----- Create setup files
if [[ ${DO_CREATE_SETUP} -eq 1 ]]; then
  root -b -l -q ${MACRO_SETUP}"(${RUN}, \"${DATA_TOP_DIR}\")" &> ${SETUP_LOG}
  cat ${SETUP_LOG}

  if [[ (1 -ne $(grep -c " Test passed" "${SETUP_LOG}")) || (1 -ne $(grep -c " All ok " "${SETUP_LOG}")) ]]; then
    printf "\nSetup file creation for %s failed, stopping there\n" "${TSA_INP}"
    rm ${SETUP_LOG}
    exit 5
  fi
  rm ${SETUP_LOG}
fi

# ----- Run unpacker
if [[ ${DO_UNPACK} -eq 1 ]]; then
  # TODO: Define unpack options for different setups
  ${ONLINE_BINARY} --steps Unpack -i ${TSA_INP} -r ${RUN} -p ${ONLINE_PAR} -O DigiTimeslice -o ${OUT_DIGI} -s STS BMON TOF TRD RICH -n ${N_TS}

  root -l -b -q ${VMCWORKDIR}/macro/run/run_inspect_digi_timeslice.C"(\"${OUT_DIGI}\", \"${INP_FILE}\")" &> ${UNPACK_LOG}
  cat ${UNPACK_LOG}
  rm ${OUT_DIGI}

  if [[ 1 -ne $(grep -c "Macro finished successfully." "${UNPACK_LOG}") ]]; then
    printf "\nUnpacked data file conversion of %s failed, stopping there\n" "${TSA_INP}"
    rm ${UNPACK_LOG}
    exit 6
  fi
  rm ${UNPACK_LOG}
fi

# ----- Run reconstruction
if [[ ${DO_RECO} -eq 1 ]]; then
  UNP_FILE_ID=-1
  cp "${DATA_TOP_DIR}/${SETUP_NAME}.par.root" "${DATA_TOP_DIR}/${FILE_LABEL}.par.root"
  DATA_PREF="${DATA_TOP_DIR}"
  PARS="${RUN},${N_TS},\"${DATA_PREF}\",\"${DATA_PREF}\",${UNP_FILE_ID},${RECO_MVD},${RECO_STS},${RECO_TRD}"
  PARS="${PARS},${RECO_TRD2d},${RECO_RICH},${RECO_MUCH},${RECO_TOF},${RECO_TOFtr},${RECO_PSD},${RECO_ALI},${RECO_EvB}"
  PARS="${PARS},${RECO_CA},${RECO_QA},${RECO_FSD},\"${INP_FILE}\",${RECO_PV}"
  root -b -l -q ${MACRO_RECO}"(${PARS})" &> ${RECO_LOG}
  cat ${RECO_LOG}

  if [[ (1 -ne $(grep -c " Test passed" "${RECO_LOG}")) || (1 -ne $(grep -c " All ok " "${RECO_LOG}")) ]]; then
    printf "\nReconstruction of %s failed, stopping there\n" "${TSA_INP}"
    rm ${RECO_LOG}
    exit 7
  fi
  rm ${RECO_LOG}

  ln -s -f "${FILE_LABEL}.digi.root" "${DATA_TOP_DIR}/${FILE_LABEL}.raw.root"  # TMP for QA
  # Commented out as the output of mcbm_event_reco_L1.C is already [...].rec.root
  #  ln -s -f "${FILE_LABEL}.reco.root" "${DATA_TOP_DIR}/${FILE_LABEL}.rec.root"  # TMP for QA
fi

# ----- Run QA of reco modules
if [[ ${DO_QA_MODULE} -eq 1 ]]; then
  QA_GEO="${DATA_TOP_DIR}/${SETUP_NAME}.geo.root"
  QA_REC="${DATA_TOP_DIR}/${FILE_LABEL}.rec.root"
  ln -s -f ${QA_GEO} "${SETUP_NAME}.geo.root"
  PARS="-1,\"${QA_REC}\",\"${SETUP_NAME}\",kFALSE,${RECO_ALI}"
  root -b -l -q ${MACRO_QA_MODULE}"(${PARS})" &> ${RECO_QA_LOG}
  cat ${RECO_QA_LOG}

  if [[ (1 -ne $(grep -c " Test passed" "${RECO_QA_LOG}")) || (1 -ne $(grep -c " All ok " "${RECO_QA_LOG}")) ]]; then
    printf "\nReco Modules QA for %s failed, stopping there\n" "${TSA_INP}"
    rm ${RECO_QA_LOG}
    exit 8
  fi
  rm ${RECO_QA_LOG}
fi

# ----- Run QA
if [[ ${DO_QA} -eq 1 ]]; then
  USE_MC="kFALSE"
  CONFIG=""
  BENCHMARK=""
  #PARS="1,\"${DATA_TOP_DIR}/${FILE_LABEL}\",\"${SETUP_NAME}\",${USE_MC},\"${CONFIG}\",\"${BENCHMARK}\""
  QA_RAW="${DATA_TOP_DIR}/${FILE_LABEL}.raw.root"
  QA_REC="${DATA_TOP_DIR}/${FILE_LABEL}.rec.root"
  QA_PAR="${DATA_TOP_DIR}/${FILE_LABEL}.par.root"
  QA_GEO="${DATA_TOP_DIR}/${SETUP_NAME}.geo.root"
  QA_OUT="${DATA_TOP_DIR}/${FILE_LABEL}.qa.root"

  PARS="0,\"\",\"${QA_RAW}\",\"${QA_REC}\",\"${QA_PAR}\",\"${QA_GEO}\",\"${QA_OUT}\",\"${SETUP_NAME}\""
  PARS="${PARS},${USE_MC},\"${CONFIG}\",\"${BENCHMARK}\",${RECO_ALI}"
  root -b -l -q ${MACRO_QA}"(${PARS})" &> ${QA_LOG}
  cat ${QA_LOG}

  if [[ (1 -eq $(grep -c " QA checks failed" "${QA_LOG}")) || (1 -ne $(grep -c " Test passed" "${QA_LOG}"))
        || (1 -ne $(grep -c " All ok " "${QA_LOG}")) ]]; then
    printf "\nFull QA for %s failed, stopping there\n" "${TSA_INP}"
    rm ${QA_LOG}
    exit 9
  fi
  rm ${QA_LOG}
fi

printf "Reconstruction of %s succeeded\n" "${TSA_INP}"
